'Secretary: a utility for removing personal data in text'

__version__ = '0.0.5'
__version_info__ = 'Secretary v' + __version__
__author__ = 'Erik Broman <mikroberna@gmail.com>'
__status__ = 'Alpha'
__copyright__ = 'Copyright (c) 2023- ' + __author__
